import csv
import os
import sys
import threading
import traceback

from flask import g

from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true

import chellow.computer
from chellow.models import Era, Session, Site, SiteEra, Supply
from chellow.utils import (
    csv_make_val,
    hh_format,
    hh_range,
    req_date,
    req_int,
)
from chellow.views import chellow_redirect


def content(supply_id, start_date, finish_date, user):
    caches = {}
    try:
        sess = Session()
        supply = Supply.get_by_id(sess, supply_id)

        forecast_date = chellow.computer.forecast_date()

        prev_titles = None
        running_name, finished_name = chellow.dloads.make_names(
            f"supply_virtual_bills_hh_{supply_id}.csv", user
        )
        f = open(running_name, mode="w", newline="")
        w = csv.writer(f, lineterminator="\n")

        for hh_start in hh_range(caches, start_date, finish_date):
            era = (
                sess.query(Era)
                .filter(
                    Era.supply == supply,
                    Era.start_date <= hh_start,
                    or_(Era.finish_date == null(), Era.finish_date >= hh_start),
                )
                .one()
            )

            site = (
                sess.query(Site)
                .join(SiteEra)
                .filter(SiteEra.era == era, SiteEra.is_physical == true())
                .one()
            )

            ds = chellow.computer.SupplySource(
                sess, hh_start, hh_start, forecast_date, era, True, caches
            )

            titles = ["MPAN Core", "Site Code", "Site Name", "Account", "HH Start", ""]

            output_line = [
                ds.mpan_core,
                site.code,
                site.name,
                ds.supplier_account,
                hh_format(ds.start_date),
                "",
            ]

            mop_titles = ds.contract_func(era.mop_contract, "virtual_bill_titles")()
            titles.extend(["mop-" + t for t in mop_titles])

            ds.contract_func(era.mop_contract, "virtual_bill")(ds)
            bill = ds.mop_bill
            for title in mop_titles:
                output_line.append(csv_make_val(bill.get(title, "")))
                if title in bill:
                    del bill[title]
            for k in sorted(bill.keys()):
                output_line.extend([k, csv_make_val(bill[k])])

            output_line.append("")
            dc_titles = ds.contract_func(era.dc_contract, "virtual_bill_titles")()
            titles.append("")
            titles.extend(["dc-" + t for t in dc_titles])

            ds.contract_func(era.dc_contract, "virtual_bill")(ds)
            bill = ds.dc_bill
            for title in dc_titles:
                output_line.append(csv_make_val(bill.get(title, "")))
                if title in bill:
                    del bill[title]

            for k in sorted(bill.keys()):
                output_line.extend([k, csv_make_val(bill[k])])

            if era.imp_supplier_contract is not None:
                contract = era.imp_supplier_contract
                output_line.append("")
                supplier_titles = ds.contract_func(contract, "virtual_bill_titles")()
                titles.append("")
                titles.extend(["imp-supplier-" + t for t in supplier_titles])

                ds.contract_func(contract, "virtual_bill")(ds)
                bill = ds.supplier_bill
                for title in supplier_titles:
                    output_line.append(csv_make_val(bill.get(title, "")))
                    if title in bill:
                        del bill[title]

                for k in sorted(bill.keys()):
                    output_line.extend([k, csv_make_val(bill[k])])

            if era.exp_supplier_contract is not None:
                contract = era.exp_supplier_contract
                ds = chellow.computer.SupplySource(
                    sess, hh_start, hh_start, forecast_date, era, False, caches
                )
                output_line.append("")
                supplier_titles = ds.contract_func(contract, "virtual_bill_titles")()
                titles.append("")
                titles.extend(["exp-supplier-" + t for t in supplier_titles])

                ds.contract_func(contract, "virtual_bill")(ds)
                bill = ds.supplier_bill
                for title in supplier_titles:
                    output_line.append(csv_make_val(bill.get(title, "")))
                    if title in bill:
                        del bill[title]

                for k in sorted(bill.keys()):
                    output_line.extend([k, csv_make_val(bill[k])])

            if titles != prev_titles:
                prev_titles = titles
                w.writerow(titles)
            w.writerow(output_line)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        w.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    supply_id = req_int("supply_id")
    start_date = req_date("start")
    finish_date = req_date("finish")

    args = supply_id, start_date, finish_date, g.user
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import csv
import os
import threading
import traceback
from datetime import datetime as Datetime, timedelta as Timedelta
from io import BytesIO, StringIO
from zipfile import ZipFile

from flask import g, render_template, request

from sqlalchemy import null
from sqlalchemy.orm import joinedload

from werkzeug.exceptions import BadRequest

import chellow.dloads
from chellow.models import (
    Llfc,
    MarketRole,
    MeterType,
    Mtc,
    Participant,
    Party,
    Session,
    VoltageLevel,
)
from chellow.utils import ct_datetime, hh_format, to_ct, to_utc
from chellow.views import chellow_redirect


def parse_date(date_str):
    if len(date_str) == 0:
        return None
    else:
        return to_utc(to_ct(Datetime.strptime(date_str, "%d/%m/%Y")))


def parse_to_date(date_str):
    if len(date_str) == 0:
        return None
    else:
        dt = to_ct(Datetime.strptime(date_str, "%d/%m/%Y"))
        dt += Timedelta(hours=23, minutes=30)
        return to_utc(dt)


def is_common_mtc(code):
    return 499 < code < 510 or 799 < code < 1000


def do_get(sess):
    return render_template("report_163.html")


VOLTAGE_MAP = {"24": {"602": {to_utc(ct_datetime(2010, 4, 1)): "LV"}}}


def _parse_Line_Loss_Factor_Class(sess, csv_reader):
    VOLTAGE_LEVEL_CODES = set([v.code for v in sess.query(VoltageLevel)])
    DNO_MAP = dict(
        (dno.participant.code, dno)
        for dno in sess.query(Party)
        .join(MarketRole)
        .filter(MarketRole.code == "R")
        .options(joinedload(Party.participant))
    )

    rows = []
    for values in csv_reader:
        participant_code = values[0]
        # market_role_code = values[1]
        llfc_code = values[3].zfill(3)
        valid_from = parse_date(values[4])
        description = values[5]
        is_import = values[6] in ("A", "B")
        is_substation = any(
            p in description for p in ("_SS", " SS", " S/S", "(S/S)", "sub", "Sub")
        )

        valid_to = parse_to_date(values[7])

        try:
            dno = DNO_MAP[participant_code]
        except KeyError:
            rows.append(("# There is no DNO with participant code ", participant_code))
            continue

        try:
            voltage_level_code = VOLTAGE_MAP[dno.dno_code][llfc_code][valid_from]
        except KeyError:
            voltage_level_code = "LV"
            description_upper = description.upper()
            for vl_code in VOLTAGE_LEVEL_CODES:
                if vl_code in description_upper:
                    voltage_level_code = vl_code
                    break

        llfc = (
            sess.query(Llfc)
            .filter(
                Llfc.dno == dno, Llfc.code == llfc_code, Llfc.valid_from == valid_from
            )
            .first()
        )

        if llfc is None:
            row = (
                "insert",
                "llfc",
                dno.dno_code,
                llfc_code,
                description,
                voltage_level_code,
                is_substation,
                is_import,
                hh_format(valid_from),
                hh_format(valid_to, ongoing_str=""),
            )
            rows.append(row)

        elif any(
            (
                description != llfc.description,
                voltage_level_code != llfc.voltage_level.code,
                is_substation != llfc.is_substation,
                is_import != llfc.is_import,
                valid_to != llfc.valid_to,
            )
        ):
            row = (
                "update",
                "llfc",
                dno.dno_code,
                llfc.code,
                hh_format(llfc.valid_from),
                description,
                voltage_level_code,
                is_substation,
                is_import,
                hh_format(valid_to, ongoing_str=""),
            )
            rows.append(row)

    return rows


def _parse_Market_Participant(sess, csv_reader):
    rows = []
    for values in csv_reader:
        participant_code = values[0]
        participant_name = values[1]

        participant = (
            sess.query(Participant).filter(Participant.code == participant_code).first()
        )

        if participant is None:
            row = "insert", "participant", participant_code, participant_name
            rows.append(row)

        elif participant_name != participant.name:
            row = "update", "participant", participant_code, participant_name
            rows.append(row)

    return rows


def _parse_Market_Role(sess, csv_reader):
    rows = []
    for values in csv_reader:
        role_code = values[0]
        role_description = values[1]

        role = sess.query(MarketRole).filter(MarketRole.code == role_code).first()

        if role is None:
            row = "insert", "market_role", role_code, role_description
            rows.append(row)

        elif role_description != role.description:
            row = "update", "market_role", role_code, role_description
            rows.append(row)

    return rows


def _parse_Market_Participant_Role(sess, csv_reader):
    rows = []
    for values in csv_reader:
        participant_code = values[0]
        market_role_code = values[1]
        valid_from = parse_date(values[2])
        party = (
            sess.query(Party)
            .join(Participant)
            .join(MarketRole)
            .filter(
                Party.valid_from == valid_from,
                Participant.code == participant_code,
                MarketRole.code == market_role_code,
            )
            .first()
        )
        valid_to = parse_to_date(values[3])
        name = values[4]
        dno_code_str = values[14]
        dno_code = None if len(dno_code_str) == 0 else dno_code_str
        if dno_code == "99":
            continue

        if party is None:
            row = (
                "insert",
                "party",
                market_role_code,
                participant_code,
                name,
                hh_format(valid_from),
                hh_format(valid_to, ongoing_str=""),
                dno_code_str,
            )
            rows.append(row)

        elif any(
            (name != party.name, dno_code != party.dno_code, valid_to != party.valid_to)
        ):
            row = (
                "update",
                "party",
                market_role_code,
                participant_code,
                name,
                hh_format(valid_from),
                hh_format(valid_to, ongoing_str=""),
                dno_code_str,
            )
            rows.append(row)

    return rows


def _parse_Meter_Timeswitch_Class(sess, csv_reader):
    rows = []
    for values in csv_reader:
        code = values[0].zfill(3)
        valid_from = parse_date(values[1])
        valid_to = parse_to_date(values[2])
        description = values[3]
        is_common = values[4] == "T"
        has_related_metering = values[5] == "T"
        meter_type_code = values[6]
        meter_payment_type_code = values[7]
        has_comms = values[8] == "T"
        is_hh = values[9] == "H"
        tpr_count_str = values[10]
        tpr_count = 0 if tpr_count_str == "" else int(tpr_count_str)

        if is_common:
            mtc = (
                sess.query(Mtc)
                .filter(
                    Mtc.dno == null(), Mtc.code == code, Mtc.valid_from == valid_from
                )
                .first()
            )
            if mtc is None:
                row = (
                    "insert",
                    "mtc",
                    "",
                    code,
                    description,
                    has_related_metering,
                    has_comms,
                    is_hh,
                    meter_type_code,
                    meter_payment_type_code,
                    tpr_count,
                    hh_format(valid_from),
                    hh_format(valid_to, ongoing_str=""),
                )
                rows.append(row)

            elif any(
                (
                    description != mtc.description,
                    has_related_metering != mtc.has_related_metering,
                    has_comms != mtc.has_comms,
                    is_hh != mtc.is_hh,
                    meter_type_code != mtc.meter_type.code,
                    meter_payment_type_code != mtc.meter_payment_type.code,
                    tpr_count != mtc.tpr_count,
                    valid_to != mtc.valid_to,
                )
            ):
                row = (
                    "update",
                    "mtc",
                    "",
                    mtc.code,
                    description,
                    has_related_metering,
                    has_comms,
                    is_hh,
                    meter_type_code,
                    meter_payment_type_code,
                    tpr_count,
                    hh_format(mtc.valid_from),
                    hh_format(valid_to, ongoing_str=""),
                )
                rows.append(row)

    return rows


def _parse_MTC_in_PES_Area(sess, csv_reader):
    rows = []
    dnos = dict(
        (p.participant.code, (p.id, p.dno_code))
        for p in sess.query(Party)
        .join(Participant)
        .join(MarketRole)
        .filter(MarketRole.code == "R")
        .options(joinedload(Party.participant))
    )
    mtcs = dict(
        ((m.dno_id, m.code, m.valid_from), m)
        for m in sess.query(Mtc)
        .options(joinedload(Mtc.meter_type), joinedload(Mtc.meter_payment_type))
        .all()
    )

    for values in csv_reader:
        code_str = values[0]
        if not Mtc.has_dno(code_str):
            continue

        code_int = int(code_str)
        code = code_str.zfill(3)
        participant_code = values[2]
        dno_id, dno_code = dnos[participant_code]
        valid_from = parse_date(values[3])
        valid_to = parse_to_date(values[4])
        description = values[5]
        meter_type_code = values[6]
        meter_payment_type_code = values[7]
        has_related_metering = code_int > 500
        has_comms = values[8] == "Y"
        is_hh = values[9] == "H"
        tpr_count_str = values[10]
        tpr_count = 0 if tpr_count_str == "" else int(tpr_count_str)

        mtc = mtcs.get((dno_id, code, valid_from))

        if mtc is None:
            row = (
                "insert",
                "mtc",
                dno_code,
                code,
                description,
                has_related_metering,
                has_comms,
                is_hh,
                meter_type_code,
                meter_payment_type_code,
                tpr_count,
                hh_format(valid_from),
                hh_format(valid_to, ongoing_str=""),
            )
            rows.append(row)

        elif any(
            (
                description != mtc.description,
                has_related_metering != mtc.has_related_metering,
                has_comms != mtc.has_comms,
                is_hh != mtc.is_hh,
                meter_type_code != mtc.meter_type.code,
                meter_payment_type_code != mtc.meter_payment_type.code,
                tpr_count != mtc.tpr_count,
                valid_to != mtc.valid_to,
            )
        ):
            row = (
                "update",
                "mtc",
                mtc.dno.dno_code,
                mtc.code,
                description,
                has_related_metering,
                has_comms,
                is_hh,
                meter_type_code,
                meter_payment_type_code,
                tpr_count,
                hh_format(mtc.valid_from),
                hh_format(valid_to, ongoing_str=""),
            )
            rows.append(row)

    return rows


def _parse_MTC_Meter_Type(sess, csv_reader):
    rows = []
    for values in csv_reader:
        code = values[0]
        description = values[1]
        valid_from = parse_date(values[2])
        valid_to = parse_to_date(values[3])
        pt = (
            sess.query(MeterType)
            .filter(MeterType.code == code, MeterType.valid_from == valid_from)
            .first()
        )
        if pt is None:
            row = (
                "insert",
                "meter_type",
                code,
                description,
                hh_format(valid_from),
                hh_format(valid_to, ongoing_str=""),
            )
            rows.append(row)

        elif (description, valid_from, valid_to) != (
            pt.description,
            pt.valid_from,
            pt.valid_to,
        ):
            row = (
                "update",
                "meter_type",
                code,
                description,
                hh_format(valid_from),
                hh_format(valid_to),
            )
            rows.append(row)

    return rows


def content(fin, user):
    sess = f = w = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            "mdd__general_import.csv", user
        )
        f = open(running_name, mode="w", newline="")
        w = csv.writer(f, lineterminator="\n")

        zip_file = ZipFile(fin)

        for zname in zip_file.namelist():
            csv_file = StringIO(zip_file.read(zname).decode("utf-8"))
            csv_reader = iter(csv.reader(csv_file))
            next(csv_reader)  # Skip titles
            rows = []

            if zname.startswith("Line_Loss_Factor_Class"):
                rows = _parse_Line_Loss_Factor_Class(sess, csv_reader)

            elif zname.startswith("Market_Participant_Role"):
                rows = _parse_Market_Participant_Role(sess, csv_reader)

            elif zname.startswith("Market_Participant"):
                rows = _parse_Market_Participant(sess, csv_reader)

            elif zname.startswith("Market_Role"):
                rows = _parse_Market_Role(sess, csv_reader)

            elif zname.startswith("Meter_Timeswitch_Class"):
                rows = _parse_Meter_Timeswitch_Class(sess, csv_reader)

            elif zname.startswith("MTC_in_PES_Area"):
                rows = _parse_MTC_in_PES_Area(sess, csv_reader)

            elif zname.startswith("MTC_Meter_Type"):
                rows = _parse_MTC_Meter_Type(sess, csv_reader)

            w.writerows(rows)

    except BaseException as e:
        if w is None:
            raise e
        else:
            w.writerow([traceback.format_exc()])

    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_post(sess):
    file_item = request.files["file"]
    file_path = file_item.filename
    file_head, file_name = os.path.split(file_path)
    file_title, file_ext = os.path.splitext(file_name)
    if file_ext != ".zip":
        raise BadRequest(
            f"The file name should have the extension .zip, but in fact it "
            f"has the extension '{file_ext}'."
        )
    f = BytesIO(file_item.read())
    f.seek(0)

    threading.Thread(target=content, args=(f, g.user)).start()
    return chellow_redirect("/downloads", 303)



import csv
import os
import sys
import threading
import traceback

from flask import g

import requests

from sqlalchemy import select
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null, or_

from werkzeug.exceptions import BadRequest

import chellow.dloads
from chellow.models import Contract, Era, Mtc, Party, Session, Source, Supply
from chellow.utils import req_bool
from chellow.views import chellow_redirect


def content(user, show_ignored):
    sess = f = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            "ecoes_comparison.csv", user
        )
        f = open(running_name, mode="w", newline="")

        props = Contract.get_non_core_by_name(sess, "configuration").make_properties()

        ECOES_KEY = "ecoes"
        try:
            ecoes_props = props[ECOES_KEY]
        except KeyError:
            raise BadRequest(
                f"The property {ECOES_KEY} cannot be found in the configuration "
                f"properties."
            )

        for key in (
            "user_name",
            "password",
            "prefix",
            "exclude_mpan_cores",
            "ignore_mpan_cores_msn",
        ):
            try:
                ecoes_props[key]
            except KeyError:
                raise BadRequest(
                    f"The property {key} cannot be found in the 'ecoes' section of "
                    f"the configuration properties."
                )

        exclude_mpan_cores = ecoes_props["exclude_mpan_cores"]
        ignore_mpan_cores_msn = ecoes_props["ignore_mpan_cores_msn"]
        url_prefix = ecoes_props["prefix"]

        proxies = props.get("proxies", {})
        s = requests.Session()
        s.verify = False
        r = s.get(url_prefix, proxies=proxies)
        data = {
            "Username": ecoes_props["user_name"],
            "Password": ecoes_props["password"],
        }
        r = s.post(url_prefix, data=data, allow_redirects=False)
        r = s.get(
            f"{url_prefix}NonDomesticCustomer/ExportPortfolioMPANs?fileType=csv",
            proxies=proxies,
        )

        _process(
            sess,
            r.text.splitlines(True),
            exclude_mpan_cores,
            ignore_mpan_cores_msn,
            f,
            show_ignored,
        )

    except BaseException as e:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        f.write(msg)
        raise e
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def _process(
    sess, ecoes_lines, exclude_mpan_cores, ignore_mpan_cores_msn, f, show_ignored
):
    writer = csv.writer(f, lineterminator="\n")

    mpans = []

    for exp in (Era.imp_mpan_core, Era.exp_mpan_core):
        for (v,) in sess.execute(
            select(exp)
            .join(Supply)
            .join(Source)
            .join(Supply.dno)
            .filter(
                Party.dno_code.notin_(("88", "99")),
                Era.finish_date == null(),
                Source.code != "3rd-party",
                exp.notin_(exclude_mpan_cores),
                exp != null(),
            )
            .distinct()
            .order_by(exp)
        ):
            mpans.append(v)

    ecoes_mpans = {}

    parser = iter(csv.reader(ecoes_lines))
    next(parser)  # Skip titles

    for values in parser:
        ecoes_titles = [
            "mpan-core",
            "address-line-1",
            "address-line-2",
            "address-line-3",
            "address-line-4",
            "address-line-5",
            "address-line-6",
            "address-line-7",
            "address-line-8",
            "address-line-9",
            "post-code",
            "supplier",
            "registration-from",
            "mtc",
            "mtc-date",
            "llfc",
            "llfc-from",
            "pc",
            "ssc",
            "measurement-class",
            "energisation-status",
            "da",
            "dc",
            "mop",
            "mop-appoint-date",
            "gsp-group",
            "gsp-effective-from",
            "dno",
            "msn",
            "meter-install-date",
            "meter-type",
            "map-id",
        ]

        ecoes_row = dict(zip(ecoes_titles, map(str.strip, values)))
        mpan_core = ecoes_row["mpan-core"]
        mpan_spaces = " ".join(
            (
                mpan_core[:2],
                mpan_core[2:6],
                mpan_core[6:10],
                mpan_core[-3:],
            )
        )
        if mpan_spaces in exclude_mpan_cores:
            continue

        ecoes_row["mpan_spaces"] = mpan_spaces
        if mpan_spaces in ecoes_mpans:
            prev_row = ecoes_mpans[mpan_spaces]
            prev_row["meter_count"] += 1
        else:
            ecoes_row["meter_count"] = 1
            ecoes_mpans[mpan_spaces] = ecoes_row

    titles = (
        "MPAN Core",
        "MPAN Core No Spaces",
        "ECOES PC",
        "Chellow PC",
        "ECOES MTC",
        "Chellow MTC",
        "ECOES LLFC",
        "Chellow LLFC",
        "ECOES SSC",
        "Chellow SSC",
        "ECOES Energisation Status",
        "Chellow Energisation Status",
        "ECOES Supplier",
        "Chellow Supplier",
        "ECOES DC",
        "Chellow DC",
        "ECOES MOP",
        "Chellow MOP",
        "ECOES GSP Group",
        "Chellow GSP Group",
        "ECOES MSN",
        "Chellow MSN",
        "ECOES Meter Type",
        "Chellow Meter Type",
        "Ignored",
        "Problem",
    )
    writer.writerow(titles)

    for mpan_spaces, ecoes in sorted(ecoes_mpans.items()):
        problem = ""
        ignore = True

        try:
            ecoes_es = ecoes["energisation-status"]
        except KeyError as e:
            raise e

        ecoes_disconnected = ecoes_es == ""
        current_chell = mpan_spaces in mpans

        if ecoes["meter_count"] > 1:
            problem += (
                f"There are {ecoes['meter_count']} meters associated with this MPAN "
                f"core in ECOES, but Chellow only supports one meter per supply at the "
                f"moment. If there really should be multiple meters for this supply, "
                f"let me know and I'll add support for it in Chellow."
            )
            ignore = False

        if ecoes_disconnected and current_chell:
            problem += "Disconnected in ECOES, but current in Chellow. "
            ignore = False

        elif not ecoes_disconnected and not current_chell:
            problem += f"In ECOES (as {ecoes_es}) but disconnected in Chellow. "
            ignore = False

        if current_chell:
            mpans.remove(mpan_spaces)
            era = sess.execute(
                select(Era)
                .filter(
                    Era.finish_date == null(),
                    or_(
                        Era.imp_mpan_core == mpan_spaces,
                        Era.exp_mpan_core == mpan_spaces,
                    ),
                )
                .options(
                    joinedload(Era.supply).joinedload(Supply.gsp_group),
                    joinedload(Era.mop_contract)
                    .joinedload(Contract.party)
                    .joinedload(Party.participant),
                    joinedload(Era.dc_contract)
                    .joinedload(Contract.party)
                    .joinedload(Party.participant),
                    joinedload(Era.imp_supplier_contract)
                    .joinedload(Contract.party)
                    .joinedload(Party.participant),
                    joinedload(Era.exp_supplier_contract)
                    .joinedload(Contract.party)
                    .joinedload(Party.participant),
                    joinedload(Era.pc),
                    joinedload(Era.imp_llfc),
                    joinedload(Era.exp_llfc),
                    joinedload(Era.mtc).joinedload(Mtc.meter_type),
                    joinedload(Era.ssc),
                    joinedload(Era.energisation_status),
                    joinedload(Era.channels),
                )
            ).scalar()

            chellow_es = era.energisation_status.code
            if ecoes_es != chellow_es:
                problem += "The energisation statuses don't match. "
                ignore = False

            if not (ecoes_es == "D" and chellow_es == "D"):
                if era.imp_mpan_core == mpan_spaces:
                    supplier_contract = era.imp_supplier_contract
                    llfc = era.imp_llfc
                else:
                    supplier_contract = era.exp_supplier_contract
                    llfc = era.exp_llfc

                chellow_pc = era.pc.code
                try:
                    if int(ecoes["pc"]) != int(chellow_pc):
                        problem += "The PCs don't match. "
                        ignore = False
                except ValueError:
                    problem += "Can't parse the PC. "
                    ignore = False

                chellow_mtc = era.mtc.code
                try:
                    if int(ecoes["mtc"]) != int(chellow_mtc):
                        problem += "The MTCs don't match. "
                        ignore = False
                except ValueError:
                    problem += "Can't parse the MTC. "
                    ignore = False

                chellow_llfc = llfc.code
                if ecoes["llfc"].zfill(3) != chellow_llfc:
                    problem += "The LLFCs don't match. "
                    ignore = False

                chellow_ssc = era.ssc
                if chellow_ssc is None:
                    chellow_ssc = ""
                    chellow_ssc_int = None
                else:
                    chellow_ssc = chellow_ssc.code
                    chellow_ssc_int = int(chellow_ssc)

                if len(ecoes["ssc"]) > 0:
                    ecoes_ssc_int = int(ecoes["ssc"])
                else:
                    ecoes_ssc_int = None

                if ecoes_ssc_int != chellow_ssc_int and not (
                    ecoes_ssc_int is None and chellow_ssc_int is None
                ):
                    problem += "The SSCs don't match. "
                    ignore = False

                chellow_supplier = supplier_contract.party.participant.code
                if chellow_supplier != ecoes["supplier"]:
                    problem += "The supplier codes don't match. "
                    ignore = False

                dc_contract = era.dc_contract
                if dc_contract is None:
                    chellow_dc = ""
                else:
                    chellow_dc = dc_contract.party.participant.code

                if chellow_dc != ecoes["dc"]:
                    problem += "The DC codes don't match. "
                    ignore = False

                mop_contract = era.mop_contract
                if mop_contract is None:
                    chellow_mop = ""
                else:
                    chellow_mop = mop_contract.party.participant.code

                if chellow_mop != ecoes["mop"]:
                    problem += "The MOP codes don't match. "
                    ignore = False

                chellow_gsp_group = era.supply.gsp_group.code
                if chellow_gsp_group != ecoes["gsp-group"]:
                    problem += "The GSP group codes don't match. "
                    ignore = False

                chellow_msn = era.msn
                if chellow_msn is None:
                    chellow_msn = ""

                if chellow_msn != ecoes["msn"]:
                    problem += "The meter serial numbers don't match. "
                    if mpan_spaces not in ignore_mpan_cores_msn:
                        ignore = False
                elif mpan_spaces in ignore_mpan_cores_msn:
                    problem += (
                        "This MPAN core is in mpan_cores_ignore and yet the meter "
                        "serial numbers do match. "
                    )

                chellow_meter_type = _meter_type(era)

                if chellow_meter_type != ecoes["meter-type"]:
                    problem += (
                        "The meter types don't match. See "
                        "https://dtc.mrasco.com/DataItem.aspx?ItemCounter=0483 "
                    )
                    ignore = False
        else:
            chellow_pc = ""
            chellow_mtc = ""
            chellow_llfc = ""
            chellow_ssc = ""
            chellow_es = ""
            chellow_supplier = ""
            chellow_dc = ""
            chellow_mop = ""
            chellow_gsp_group = ""
            chellow_msn = ""
            chellow_meter_type = ""

        if len(problem) > 0 and not (not show_ignored and ignore):
            writer.writerow(
                [
                    mpan_spaces,
                    ecoes["mpan-core"],
                    ecoes["pc"],
                    chellow_pc,
                    ecoes["mtc"],
                    chellow_mtc,
                    ecoes["llfc"],
                    chellow_llfc,
                    ecoes["ssc"],
                    chellow_ssc,
                    ecoes["energisation-status"],
                    chellow_es,
                    ecoes["supplier"],
                    chellow_supplier,
                    ecoes["dc"],
                    chellow_dc,
                    ecoes["mop"],
                    chellow_mop,
                    ecoes["gsp-group"],
                    chellow_gsp_group,
                    ecoes["msn"],
                    chellow_msn,
                    ecoes["meter-type"],
                    chellow_meter_type,
                    ignore,
                    problem,
                ]
            )
        sess.expunge_all()

    for mpan_core in mpans:
        supply = Supply.get_by_mpan_core(sess, mpan_core)
        era = supply.find_era_at(sess, None)
        if era.imp_mpan_core == mpan_core:
            supplier_contract = era.imp_supplier_contract
            llfc = era.imp_llfc
        else:
            supplier_contract = era.exp_supplier_contract
            llfc = era.exp_llfc

        ssc = "" if era.ssc is None else era.ssc.code

        es = era.energisation_status.code

        dc_contract = era.dc_contract
        dc = "" if dc_contract is None else dc_contract.party.participant.code

        mop_contract = era.mop_contract
        mop = "" if mop_contract is None else mop_contract.party.participant.code

        msn = "" if era.msn is None else era.msn

        meter_type = _meter_type(era)

        writer.writerow(
            [
                mpan_core,
                mpan_core.replace(" ", ""),
                "",
                era.pc.code,
                "",
                era.mtc.code,
                "",
                llfc.code,
                "",
                ssc,
                "",
                es,
                "",
                supplier_contract.party.participant.code,
                "",
                dc,
                "",
                mop,
                "",
                supply.gsp_group.code,
                "",
                msn,
                "",
                meter_type,
                False,
                "In Chellow, but not in ECOES.",
            ]
        )


def _meter_type(era):
    props = era.props
    try:
        return props["meter_type"]
    except KeyError:
        if era.pc.code == "00":
            return "H"
        elif len(era.channels) > 0:
            return "RCAMR"
        elif era.mtc.meter_type.code in ["UM", "PH"]:
            return ""
        else:
            return "N"


def do_get(sess):
    show_ignored = req_bool("show_ignored")
    threading.Thread(target=content, args=(g.user, show_ignored)).start()
    return chellow_redirect("/downloads", 303)
